import numpy as np
import os.path
from os.path import join as pjoin
from refnx.analysis import CurveFitter
from refellips import (
    RI,
    Cauchy,
    DataSE,
    ReflectModelSE,
    ObjectiveSE,
    SlabSE,
    StructureSE,
    load_material,
)
from numpy.testing import assert_allclose

pth = os.path.dirname(os.path.abspath(__file__))


def test_bare_against_wvase1():
    # Check bare interface ellipsometry calculation against wvase
    dname = pjoin(pth, "TestData_bareSI.txt")
    data = DataSE(data=dname)

    si = load_material("silicon")
    void = load_material("void")

    struc = void() | si()
    assert isinstance(struc[0], SlabSE)
    assert isinstance(struc, StructureSE)

    model = ReflectModelSE(struc)

    wav, aois, psi_d, delta_d = data.data
    wavelength_aoi = np.c_[wav, aois]
    psi, delta = model(wavelength_aoi)

    assert_allclose(psi, psi_d, rtol=0.002)
    assert_allclose(delta, delta_d, rtol=0.003)

    _ = struc.ri_profile()


def test_Cauchy_against_wvase2():
    # Check Cauchy model against wvase
    A = 1.47
    B = 0.00495
    C = 0

    cauchy = Cauchy(A=A, B=B, C=C)

    wvase_output = np.loadtxt(pjoin(pth, "Cauchynk_fromWVASE.txt"))
    wavs = wvase_output[:, 0]

    refin = A + B / ((wavs / 1000) ** 2) + C / ((wavs / 1000) ** 4)

    refellips_RI = cauchy.complex(wavs)

    assert_allclose(refellips_RI, refin)
    assert_allclose(refellips_RI, wvase_output[:, 1], rtol=1e-6)


def test_refellips_against_wvase3():
    # Three layer stack(cauchy & from file) compared to wvase
    # Also tests behaviour of Delta around 180°
    dname = os.path.join(pth, "TestData_cauchy.txt")
    data = DataSE(data=dname)

    _f = os.path.join(pth, "../materials/silicon.csv")
    si = RI(_f)

    _f = os.path.join(pth, "../materials/void.csv")
    void = RI(_f)

    cauchy = Cauchy(A=1.47, B=0.00495, C=0)
    struc = void() | cauchy(1000) | si()
    assert isinstance(struc, StructureSE)

    model = ReflectModelSE(struc)
    model._flip_delta = True

    wav, aois, psi_d, delta_d = data.data
    wavelength_aoi = np.c_[wav, aois]
    psi, delta = model(wavelength_aoi)

    # assert_allclose(psi, psi_d, rtol=0.0005)
    assert_allclose(delta, delta_d, rtol=0.003)


def test_refellips_against_wvase4():
    # A 30 nm SiO film with ambient as water compared to WVASE
    dname = pjoin(pth, "WVASE_example_30nmSiO2_Water_MultiWavelength.txt")
    data = DataSE(dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    h2o = RI(pjoin(pth, "../materials/water.csv"))

    struc = h2o() | sio2(300) | si()
    struc.solvent = h2o

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, aoi])

    assert_allclose(psi, d_psi, rtol=5e-4)
    assert_allclose(delta, d_delta, rtol=5e-4)


def test_refellips_against_wvase5():
    # A 3 nm SiO film with 30 % water compared to WVASE
    dname = pjoin(pth, "WVASE_example_3nmSiO2_30EMA_MultiWavelength.txt")
    data = DataSE(dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    h2o = RI(pjoin(pth, "../materials/water.csv"))

    silica = sio2(30)
    silica.name = "Silica"
    silica.vfsolv.setp(value=0.3)

    struc = h2o() | silica | si()
    struc.solvent = h2o
    struc.ema = "linear"

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, aoi])

    assert_allclose(psi, d_psi, rtol=6e-4)
    assert_allclose(delta, d_delta, rtol=1e-4)


def test_refellips_against_wvase6():
    # A comparison to WVASE for a 2 nm SiO and 20 nm polymer film system
    # with a 50 % volume fraction of solvent in the polymer film.
    dname = pjoin(
        pth, "WVASE_example_2nmSiO2_20nmPNIPAM_50EMA_MultiWavelength.txt"
    )
    data = DataSE(dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    polymer = Cauchy(A=1.47, B=0.00495)
    h2o = Cauchy(A=1.3242, B=0.003064)

    polymer_layer = polymer(200)
    polymer_layer.name = "PNIPAM"
    polymer_layer.vfsolv.setp(value=0.5)

    struc = h2o() | polymer_layer | sio2(20) | si()
    struc.solvent = h2o

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, aoi])

    assert_allclose(psi, d_psi, rtol=6e-4)
    assert_allclose(delta, d_delta, rtol=6e-4)


def test_refellips_against_wvase7():
    # A comparison to WVASE for a 2 nm SiO and 117 nm gold film in air.
    dname = pjoin(pth, "WVASE_example_2nmSiO2_117nmAu_MultiWavelength.txt")
    data = DataSE(dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    gold = RI(pjoin(pth, "../materials/gold.csv"))
    air = RI(pjoin(pth, "../materials/air.csv"))

    struc = air() | gold(1170) | sio2(20) | si()

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True
    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, aoi])

    assert_allclose(psi, d_psi, rtol=3e-4)
    assert_allclose(delta, d_delta, rtol=3e-4)


def test_refellips_against_wvase8():
    # A comparison to WVASE for a 2 nm SiO and
    # 65 nm aluminium oxide film in air.
    dname = pjoin(pth, "WVASE_example_2nmSiO2_65nmAl2O3_MultiWavelength.txt")
    data = DataSE(dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    al2o3 = RI(pjoin(pth, "../materials/aluminium_oxide.csv"))
    air = RI(pjoin(pth, "../materials/air.csv"))

    struc = air() | al2o3(650) | sio2(20) | si()

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, aoi])

    assert_allclose(psi, d_psi, rtol=3e-4)
    assert_allclose(delta, d_delta, rtol=3e-4)


def test_refellips_against_wvase9():
    # A comparison to WVASE for a 10 nm SiO2 and
    # 325 nm aluminium oxide film in air.
    dname = pjoin(pth, "WVASE_example_10nmSiO2_325nmAl2O3_MultiWavelength.txt")
    data = DataSE(dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    al2o3 = RI(pjoin(pth, "../materials/aluminium_oxide.csv"))
    air = RI(pjoin(pth, "../materials/air.csv"))

    struc = air() | al2o3(3250) | sio2(100) | si()

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, aoi])

    assert_allclose(psi, d_psi, rtol=5e-5)
    assert_allclose(delta, d_delta, rtol=3e-4)


def test_refellips_against_wvase10():
    # A comparison to WVASE for a 3 nm SiO$_2$ and 90 nm
    # polymer film in water with 20% solvent using the
    # Maxwell-Garnett EMA method.
    dname = pjoin(
        pth, "WVASE_example_3nmSiO2_90nmPNIPAM_20EMA-MG_MultiWavelength.txt"
    )
    data = DataSE(dname)

    si = load_material("silicon")
    sio2 = load_material("silica")
    polymer = Cauchy(A=1.47, B=0.00495)
    water = Cauchy(A=1.3242, B=0.003064)

    polymer_layer = polymer(900)
    polymer_layer.name = "PNIPAM"
    polymer_layer.vfsolv.setp(value=0.2)

    struc = water() | polymer_layer | sio2(30) | si()
    struc.solvent = water
    struc.ema = "maxwell-garnett"

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, np.ones_like(wavelength) * aoi])

    assert_allclose(psi, d_psi, rtol=7e-4)
    assert_allclose(delta, d_delta, rtol=4e-4)


def test_refellips_against_wvase11():
    # A comparison to WVASE for a 6 nm SiO2 and 145 nm
    # polymer film in water with 70% solvent using the
    # Bruggeman EMA method with a depolarisation factor
    # of 0.2.
    dname = pjoin(
        pth, "WVASE_example_6nmSiO2_145nmPolymer_70EMA-BG_MultiWavelength.txt"
    )
    data = DataSE(dname)

    si = load_material("silicon")
    sio2 = load_material("silica")
    polymer = Cauchy(A=1.66, B=0.006)
    water = load_material("water")

    polymer_layer = polymer(1450)
    polymer_layer.name = "PNIPAM"
    polymer_layer.vfsolv.setp(value=0.70)

    struc = water() | polymer_layer | sio2(60) | si()
    struc.solvent = water
    struc.ema = "bruggeman"
    struc.depolarisation_factor = 0.2

    model = ReflectModelSE(struc, delta_offset=0)
    model._flip_delta = True

    wavelength, aoi, d_psi, d_delta = data.data
    psi, delta = model(np.c_[wavelength, np.ones_like(wavelength) * aoi])

    assert_allclose(psi, d_psi, rtol=2e-4)
    assert_allclose(delta, d_delta, rtol=7e-5)


def test_smoke_test_a_fit():
    dname = pjoin(pth, "WVASE_example_2nmSiO2_20nmPNIPAM_MultiWavelength.txt")
    data = DataSE(data=dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    PNIPAM = RI(pjoin(pth, "../materials/pnipam.csv"))
    air = RI(pjoin(pth, "../materials/air.csv"))

    PNIPAM_layer = PNIPAM(150)
    PNIPAM_layer.thick.setp(vary=True, bounds=(100, 500))

    struc = air() | PNIPAM_layer | sio2(20) | si()
    model = ReflectModelSE(struc)

    objective = ObjectiveSE(model, data)
    fitter = CurveFitter(objective)
    fitter.fit(method="least_squares")
    assert objective.chisqr() < 0.055


def test_logl():
    dname = pjoin(pth, "WVASE_example_2nmSiO2_20nmPNIPAM_MultiWavelength.txt")
    data = DataSE(data=dname)

    si = RI(pjoin(pth, "../materials/silicon.csv"))
    sio2 = RI(pjoin(pth, "../materials/silica.csv"))
    PNIPAM = RI(pjoin(pth, "../materials/pnipam.csv"))
    air = RI(pjoin(pth, "../materials/air.csv"))

    PNIPAM_layer = PNIPAM(150)
    PNIPAM_layer.thick.setp(vary=True, bounds=(100, 500))

    struc = air() | PNIPAM_layer | sio2(20) | si()
    model = ReflectModelSE(struc)

    objective = ObjectiveSE(model, data)
    assert_allclose(objective.logl() / -0.5, objective.chisqr())



# -*- coding: utf-8 -*-

""""
A basic representation of a 1D dataset
"""

import numpy as np
import pandas as pd
from refnx._lib import possibly_open_file

pd.options.mode.chained_assignment = None


class DataSE(object):
    r"""
    A basic representation of a 1D dataset.

    Parameters
    ----------
    data : str, file-like or tuple of np.ndarray, optional
        String pointing to a data file.
        Alternatively it is a tuple containing the data from which the dataset
        will be constructed. The tuple should have 4 members.

            - data[0] - Wavelength (nm)
            - data[1] - Angle of incidence (degree)
            - data[2] - Psi
            - data[3] - Delta

        `data` must be four long.
        All arrays must have the same shape.

    mask : array-like
        Specifies which data points are (un)masked. Must be broadcastable
        to the data. `Data1D.mask = None` clears the mask. If a mask value
        equates to `True`, then the point is included, if a mask value equates
        to `False` it is excluded.

    reflect_delta : bool
        Specifies whether delta values are reflected around 180 degrees
        (i.e., 360 - delta[delta > 180]), as is standard for some ellipsometry
        analysis packages (i.e., WVASE).

    Attributes
    ----------
    data : tuple of np.ndarray
        The data, (wavelength, AOI, psi, delta)
    finite_data : tuple of np.ndarray
        Data points that are finite
    wavelength : np.ndarray
        wavelength (nm)
    AOI : np.ndarray
        angle of incidence (degree)
    psi : np.ndarray
        psi
    delta : np.ndarray
        delta
    mask : np.ndarray
        mask
    filename : str or None
        The file the data was read from
    weighted : bool
        Whether the y data has uncertainties
    metadata : dict
        Information that should be retained with the dataset.

    """

    def __init__(
        self, data=None, name=None, delimiter="\t", reflect_delta=False, **kwds
    ):
        self.filename = None

        self.delimiter = delimiter
        self.metadata = kwds
        self._wavelength = np.zeros(0)
        self._aoi = np.zeros(0)
        self._psi = np.zeros(0)
        self._delta = np.zeros(0)
        # TODO when we come up with measurement uncertainties change this.
        self.weighted = False
        self.name = name

        # If a file, then open and load the file.
        if hasattr(data, "read") or type(data) is str:
            self.load(data)
            self.filename = data

        # If already a DataSE object, then just use that.
        elif isinstance(data, DataSE):
            self.name = data.name
            self.filename = data.filename
            self.metadata = data.metadata
            self._wavelength = data._wavelength
            self._aoi = data._aoi
            self._psi = data._psi
            self._delta = data._delta

        # If a list or tuple, then assume its in format wavelength, AOI, psi, delta.
        elif isinstance(data, (list, tuple, np.ndarray)):
            self._wavelength = data[0]
            self._aoi = data[1]
            self._psi = data[2]
            self._delta = data[3]

        self._delta_flipped = False
        if reflect_delta:
            dmask = self._delta > 180
            self._delta[dmask] = 360 - self._delta[dmask]
            self._delta_flipped = True

        self.mask = np.ones_like(self._wavelength, dtype=bool)

    def __len__(self):
        """Number of unmasked points in the dataset."""
        return self.wavelength.size

    def __str__(self):
        return "<{0}>, {1} points".format(self.name, len(self))

    def __repr__(self):
        msk = self.mask
        if np.all(self.mask):
            msk = None

        d = {"filename": self.filename, "msk": msk, "data": self.data}
        if self.filename is not None:
            return "Data1D(data={filename!r}," " mask={msk!r})".format(**d)
        else:
            return "Data1D(data={data!r}," " mask={msk!r})".format(**d)

    def unique_wavelength_data(self):
        """
        Generator yielding wavelength, AOI, psi, delta tuples for the unique
        wavelengths in a dataset (i.e. all the data points for a given
        wavelength)

        Returns
        -------
        wavelength, AOI, psi, delta
        """
        unique_wavs = np.unique(self.wavelength)
        for unique_wav in unique_wavs:
            loc = np.where(self.wavelength == unique_wav)
            yield unique_wav, self.aoi[loc], self.psi[loc], self.delta[loc]

    @property
    def wavelength(self):
        """Wavelength."""
        if self._wavelength.size > 0:
            return self._wavelength[self.mask]
        else:
            return self._wavelength

    @property
    def aoi(self):
        """Angle of incidence."""
        if self._aoi.size > 0:
            return self._aoi[self.mask]
        else:
            return self._aoi

    @property
    def psi(self):
        """Ellipsometric parameter psi."""
        if self._psi.size > 0:
            return self._psi[self.mask]
        else:
            return self._psi

    @property
    def delta(self):
        """Ellipsometric parameter delta."""
        if self._delta.size > 0:
            return self._delta[self.mask]
        else:
            return self._delta

    @property
    def data(self):
        """4-tuple containing the (lambda, AOI, psi, delta) data."""
        return self.wavelength, self.aoi, self.psi, self.delta

    @data.setter
    def data(self, data_tuple):
        """
        Set the data for this object from the supplied data.

        Parameters
        ----------
        data_tuple : tuple
            4 member tuple containing the (wav, aoi, psi, delta) data to
            specify the dataset.

        Notes
        -----
        Clears the mask for the dataset, it will need to be reapplied.

        """
        self._wavelength = np.array(data_tuple[0], dtype=float)
        self._aoi = np.array(data_tuple[1], dtype=float)
        self._psi = np.array(data_tuple[2], dtype=float)
        self._delta = np.array(data_tuple[3], dtype=float)

        self.mask = np.ones_like(self._wavelength, dtype=bool)

    def save(self, f):
        """
        Save the data to file. Saves the data as a 4 column ASCII file.

        Parameters
        ----------
        f : file-handle or string
            File to save the dataset to.

        """
        header = "wavelength\tAOI\tPsi\tDelta"
        np.savetxt(
            f,
            np.column_stack(
                (self._wavelength, self._aoi, self._psi, self._delta)
            ),
            delimiter="\t",
            header=header,
        )

    def load(self, f):
        """
        Load a dataset from file.
        Must be a 4 column ASCII file with columns [wavelength, AOI, Psi, Delta].

        Parameters
        ----------
        f : file-handle or string
            File to load the dataset from.

        """

        skip_lines = 0
        with possibly_open_file(f, "r") as text:
            for i in range(100):  # check the first 100 lines
                try:
                    float(text.readline().split(self.delimiter)[0])
                    break
                except ValueError:
                    skip_lines += 1

        self._wavelength, self._aoi, self._psi, self._delta = np.loadtxt(
            f, skiprows=skip_lines, delimiter=self.delimiter, encoding="utf8"
        ).T

    def refresh(self):
        """
        Refreshes a previously loaded dataset.

        """
        if self.filename is not None:
            with open(self.filename) as f:
                self.load(f)


def open_EP4file(fname, reflect_delta=False):
    """
    Open and load in an Accurion EP4 formmated data file.
    Typically a .dat file.

    Note: This file parser has been written for specific Accurion ellipsometers
    EP3 and EP4. No work has been done to ensure it is compatable with all
    Accurion ellipsometers. If you have trouble with this parser contact the
    maintainers through github.

    Parameters
    ----------
    fname : file-handle or string
        File to load the dataset from.

    reflect_delta : bool
        Option to reflect delta around 180 degrees (as WVASE would).

    Returns
    ----------
    datasets : DataSE structure
        Structure containing wavelength, angle of incidence, psi and delta.


    """
    df = pd.read_csv(fname, sep="\t", skiprows=[1])
    df = df.dropna(axis=0, how="any")

    try:
        df["Time"]
        time_data = True
    except KeyError:
        time_data = False
        print("No time data.")

    if time_data and len(df["Time"].drop_duplicates()) > 1:
        print("Treating as time series:")
        output = []
        for t in df["Time"].drop_duplicates():
            tdf = df[df["Time"] == t]
            output += _loadEP4(tdf)  # not sure if this will work
            output[-1]["time"] = t
    else:
        output = _loadEP4(df)
        for op in output:
            op["time"] = None

    datasets = []
    for op in output:
        data = [op["lambda"], op["aoi"], op["psi"], op["delta"]]
        del op["lambda"]
        del op["aoi"]
        del op["psi"]
        del op["delta"]
        name = _make_EP4dname(fname, op)
        datasets.append(
            DataSE(data, name=name, reflect_delta=reflect_delta, **op)
        )

    if len(datasets) == 1:
        return datasets[0]
    else:
        return datasets


def _make_EP4dname(name, metadata):
    """
    Create a helpful name for a data set based on an Accurion EP4
    formatted data file.

    Parameters
    ----------
    name : file-handle or string
        File name of data set.

    metadata : dict
        Dict containinng 'X pos', 'Y pos' and 'time' data.

    Returns
    ----------
    base : string
        Helpful name for the data set.

    """
    base = name[: -len("_20200929-083122.ds.dat")]
    if metadata["X pos"] is not None:
        base += f"_x={metadata['X pos']}mm_y={metadata['Y pos']}mm"
    if metadata["time"] is not None:
        base += f"_t={metadata['time']}s"

    return base


def custom_round(x, base=0.25):
    """
    Perform rounding to a particular base. Default base is 0.25.

    Parameters
    ----------
    x : DataFrame, array or list
        Data to be rounded.

    base : float
        Base that the rounding will be with respect to.

    Returns
    ----------
    Result of cutsom round : np.array

    """
    x = np.array(x, dtype=float)
    return np.round((base * np.round(x / base)), 2)


def _loadEP4(df):
    """
    Specifically loading a data file created by an Accurion EP4 ellipsometer.
    Dataframe should have colums ['#Lambda','AOI','Psi','Delta'].
    Optionally can also have columns [X_pos, Y_pos].


    Parameters
    ----------
    df : DataFrame
        Data frame containing the wavelength, angle of incidence, psi and
        delta data.

    Returns
    ----------
    output : list of dicts
        Dicts containing wavelength, angle of indcidence, psi, delta and
        possible X pos and Y pos.

    """

    try:
        df["X_pos"]
        df["Y_pos"]
        loc_data = True
    except KeyError:
        loc_data = False

    if loc_data and (
        len(df["X_pos"].drop_duplicates()) > 1
        or len(df["Y_pos"].drop_duplicates()) > 1
    ):
        print("Treating as multiple locations")
        df = df[["#Lambda", "AOI", "Psi", "Delta", "X_pos", "Y_pos"]]
        df.loc[:, "X_pos"] = custom_round(df["X_pos"], base=0.25)
        df.loc[:, "Y_pos"] = custom_round(df["Y_pos"], base=0.25)

        output = []
        for x in df["X_pos"].drop_duplicates():
            for y in df["Y_pos"].drop_duplicates():
                pdf = df[np.logical_and(df["X_pos"] == x, df["Y_pos"] == y)]
                pdf = pdf[["#Lambda", "AOI", "Psi", "Delta"]]
                if len(pdf.index) > 0:
                    ave_pos = pdf.groupby(["AOI", "#Lambda"]).mean()
                    ave_pos = ave_pos.reset_index()
                    summary = {
                        "lambda": np.array(ave_pos["#Lambda"]),
                        "aoi": np.array(ave_pos["AOI"]),
                        "psi": np.array(ave_pos["Psi"]),
                        "delta": np.array(ave_pos["Delta"]),
                        "X pos": x,
                        "Y pos": y,
                    }
                    output.append(summary)
    else:
        print("Treating as single location")
        df = df[["#Lambda", "AOI", "Psi", "Delta"]]
        ave_pos = df.groupby(["AOI", "#Lambda"]).mean()
        ave_pos = ave_pos.reset_index()

        summary = {
            "lambda": np.array(ave_pos["#Lambda"]),
            "aoi": np.array(ave_pos["AOI"]),
            "psi": np.array(ave_pos["Psi"]),
            "delta": np.array(ave_pos["Delta"]),
            "X pos": None,
            "Y pos": None,
        }
        output = [summary]

    return output


def open_HORIBAfile(
    fname, reflect_delta=False, lambda_cutoffs=[-np.inf, np.inf]
):
    """
    Opening and loading in a data file created by a Horiba ellipsometer. Data
    file loaded should be of the Horiba file format .spe.

    Note: This file parser has been written for a specific ellipsometer, no
    work has been done to ensure it is compatable with all Horiba
    ellipsometers. If you have trouble with this parser contact the maintainers
    through github.

    Parameters
    ----------
    fname : file-handle or string
        File to load the dataset from.

    reflect_delta : bool
        Option to reflect delta around 180 degrees (as WVASE would).

    lambda_cutoffs : list
        Specifies the minimum and maximum wavelengths of data to be loaded.
        List has length 2.

    Returns
    ----------
    DataSE : DataSE structure
        The data file structure from the loaded Horiba file.

    """

    name = fname[:-4]
    metadata = {}
    linenodict = {}
    MDingest = False

    with open(fname, "r") as f:
        lines = f.readlines()

        for i, line in enumerate(lines):
            l = line[:-1]  # Drop newline character
            if not MDingest:
                if len(l) and l[0] == "#":
                    MDlabel = " ".join(l.split(" ")[1:])[:-1]
                    metadata[MDlabel] = []
                    linenodict[MDlabel] = i
                    MDingest = True

            else:
                if not len(l):
                    MDingest = False
                    if not len(
                        metadata[MDlabel]
                    ):  # there is no metadata for entry
                        metadata[MDlabel] = None  # Set metadata to none
                    elif len(metadata[MDlabel]) == 1:  # there is only one entry
                        metadata[MDlabel] = metadata[MDlabel][
                            0
                        ]  # remove data from list

                else:  # there is metadata in the line
                    metadata[MDlabel].append(l)  # append line to metadata entry

    data_df = pd.read_csv(
        fname,
        skiprows=linenodict["DATA"] + 1,
        nrows=len(metadata["DATA"]) - 1,
        encoding="ANSI",
        delimiter=" ",
        usecols=["nm", "Psi", "Delta"],
    )

    AOI = float(metadata["INCIDENCE ANGLE"][:5])
    data_df["AOI"] = AOI * np.ones_like(data_df["nm"])
    data_df = data_df[data_df["nm"] > lambda_cutoffs[0]]
    data_df = data_df[data_df["nm"] < lambda_cutoffs[1]]

    data = [data_df["nm"], data_df["AOI"], data_df["Psi"], data_df["Delta"]]

    return DataSE(data, name=name, reflect_delta=reflect_delta, **metadata)



""""""
"""
Calculates the ellipsometry parameters Δ and Ψ from a stratified series of
layers.

TBSD 3-Clause License

Copyright (c) 2020-2022,
Hayden Robertson (University of Newcastle)
Isaac Gresham (University of Sydney)
Andrew Nelson (ANSTO)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Furthermore, here we adapt code from Byrnes (i.e. the coh_tmm function) under
the MIT license listed below:

Copyright (C) 2012-2020 Steven Byrnes

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# -*- coding: utf-8 -*-

import numpy as np
from numpy.lib.scimath import arcsin
from refellips import StructureSE

EPSILON = np.finfo(np.float64).eps

from refnx.analysis import (
    Parameters,
    Parameter,
    possibly_create_parameter,
    Transform,
)


def interface_r_s(n_i, n_f, th_i, th_f):
    return (n_i * np.cos(th_i) - n_f * np.cos(th_f)) / (
        n_i * np.cos(th_i) + n_f * np.cos(th_f)
    )


def interface_r_p(n_i, n_f, th_i, th_f):
    return (n_f * np.cos(th_i) - n_i * np.cos(th_f)) / (
        n_f * np.cos(th_i) + n_i * np.cos(th_f)
    )


def interface_t_s(n_i, n_f, th_i, th_f):
    return 2 * n_i * np.cos(th_i) / (n_i * np.cos(th_i) + n_f * np.cos(th_f))


def interface_t_p(n_i, n_f, th_i, th_f):
    return 2 * n_i * np.cos(th_i) / (n_f * np.cos(th_i) + n_i * np.cos(th_f))


def coh_tmm(n_list, d_list, th_0, lam_vac):
    """
    Code adapted by that of Byrnes - see https://arxiv.org/abs/1603.02720

    n_list is the list of refractive indices, in the order that the light would
    pass through them. The 0'th element of the list should be the semi-infinite
    medium from which the light enters, the last element should be the semi-
    infinite medium to which the light exits (if any exits).

    th_0 is the angle of incidence: 0 for normal, pi/2 for glancing.
    Remember, for a dissipative incoming medium (n_list[0] is not real), th_0
    should be complex so that n0 sin(th0) is real (intensity is constant as
    a function of lateral position).

    d_list is the list of layer thicknesses (front to back). Should correspond
    one-to-one with elements of n_list. First and last elements should be "inf".

    lam_vac is vacuum wavelength of the light.

    """
    # Convert lists to numpy arrays if they're not already.
    n_list = np.asarray(n_list)
    d_list = np.asfarray(d_list)
    num_layers = n_list.size

    th_0, lam_vac = [np.array(a) for a in np.broadcast_arrays(th_0, lam_vac)]
    orig_shp = th_0.shape
    th_0 = np.ravel(th_0)
    lam_vac = np.ravel(lam_vac)

    # Input tests
    assert (
        np.abs(np.imag(n_list[0] * np.sin(th_0))) < 100 * EPSILON
    ).all(), "Error in n0 or th0!"

    # th_list is a list with, for each layer, the angle that the light travels
    # through the layer. Computed with Snell's law. Note that the "angles" may
    # be complex!
    # Important that the arcsin here is numpy.lib.scimath.arcsin, not
    # numpy.arcsin! (They give different results e.g. for arcsin(2).)

    # (NLAYERS, NUMPNTS)
    th_list = arcsin(n_list[0] * np.sin(th_0[None, :]) / n_list[:, None])

    # kz is the z-component of (complex) angular wavevector for forward-moving
    # wave. Positive imaginary part means decaying.
    # kz_list.shape = (NLAYERS, NUMPOINTS)
    kz_list = 2 * np.pi * n_list[:, None] * np.cos(th_list) / lam_vac

    if num_layers > 2:
        # delta is the total phase accrued by traveling through a given layer.
        # don't work it out for the fronting/backing media
        # delta.shape = (NLAYERS - 2, NUMPNTS)
        delta = kz_list[1:-1] * d_list[1:-1, None]

        # For a very opaque layer, reset delta to avoid divide-by-0 and similar
        # errors. The criterion imag(delta) > 35 corresponds to single-pass
        # transmission < 1e-30 --- small enough that the exact value doesn't
        # matter.
        if (np.imag(delta) > 35).any():
            np.clip(delta.imag, -np.inf, 35, out=delta.imag)

    results = {}

    polarisations = [
        ["s", interface_r_s, interface_t_s],
        ["p", interface_r_p, interface_t_p],
    ]

    for pol, interface_r, interface_t in polarisations:
        # t_list and r_list are transmission and reflection amplitudes,
        # respectively, coming from i, going to j.
        # r_list.shape = (NLAYERS - 1, NUMPOINTS)
        r_list = interface_r(
            n_list[:-1, None], n_list[1:, None], th_list[:-1], th_list[1:]
        )
        t_list = interface_t(
            n_list[:-1, None], n_list[1:, None], th_list[:-1], th_list[1:]
        )

        # At the interface between the (n-1)st and nth material, let v_n be the
        # amplitude of the wave on the nth side heading forwards (away from the
        # boundary), and let w_n be the amplitude on the nth side heading backwards
        # (towards the boundary). Then (v_n,w_n) = M_n (v_{n+1},w_{n+1}). M_n is
        # M_list[n]. M_0 and M_{num_layers-1} are not defined.
        # My M is a bit different than Sernelius's, but Mtilde is the same.

        if num_layers > 2:
            # calculate the characteristic matrices for all the layers
            beta = np.exp(1j * delta)
            beta_inv = 1 / beta

            # M_list00.shape = (NUM_LAYERS - 2, NUMPOINTS)
            M_list00 = beta_inv / t_list[1:]
            M_list01 = r_list[1:] * beta_inv / t_list[1:]
            M_list10 = r_list[1:] * beta / t_list[1:]
            M_list11 = beta / t_list[1:]

        # initial interface
        # Mtilde00.shape = (NUMPOINTS,)
        Mtilde00 = 1 / t_list[0]
        Mtilde01 = r_list[0] / t_list[0]
        Mtilde10 = Mtilde01
        Mtilde11 = Mtilde00

        # propagate characteristic matrices
        for i in range(0, num_layers - 2):
            # matrix multiply Mtilde by characteristic matrix
            p00 = Mtilde00 * M_list00[i, :] + Mtilde01 * M_list10[i, :]
            p01 = Mtilde00 * M_list01[i, :] + Mtilde01 * M_list11[i, :]
            p10 = Mtilde10 * M_list00[i, :] + Mtilde11 * M_list10[i, :]
            p11 = Mtilde10 * M_list01[i, :] + Mtilde11 * M_list11[i, :]

            Mtilde00 = p00
            Mtilde01 = p01
            Mtilde10 = p10
            Mtilde11 = p11

        # Net complex transmission and reflection amplitudes
        r = Mtilde10 / Mtilde00
        t = 1 / Mtilde00
        results[f"r_{pol}"] = r
        results[f"t_{pol}"] = t

    return results


def Delta_Psi_TMM(AOI, layers, wavelength, delta_offset, reflect_delta=False):
    """
    Get delta and psi using the transfer matrix method.

    Parameters
    ----------
    AOI: array_like
        the angle of incidence values required for the calculation.
        Units = degrees
    Wavelength: float
        Wavelength of light. Units = nm
    layers: np.ndarray
        coefficients required for the calculation, has shape (2 + N, 4),
        where N is the number of layers
        layers[0, 1] - refractive index of fronting
        layers[0, 2] - extinction coefficent of fronting
        layers[N, 0] - thickness of layer N
        layers[N, 1] - refractive index of layer N
        layers[N, 2] - extinction coefficent of layer N
        layers[N, 3] - roughness between layer N-1/N (IGNORED!)
        layers[-1, 1] - refractive index of backing
        layers[-1, 2] - extinction coefficent of backing
        layers[-1, 3] - roughness between backing and last layer (IGNORED!)

    Returns
    -------
    Psi: np.ndarray
        Calculated Psi values for each aoi value.
    Delta: np.ndarray
        Calculated Delta values for each aoi value.


    """
    AOI = np.asfarray(AOI)
    AOI = AOI * (np.pi / 180)

    layers[0, 2] = 0  # infinate medium cannot have an extinction coeff
    RIs = layers[:, 1] + layers[:, 2] * 1j
    thicks = layers[:, 0] / 10  # Ang to nm
    thicks[0] = np.inf
    thicks[-1] = np.inf

    results = coh_tmm(n_list=RIs, d_list=thicks, th_0=AOI, lam_vac=wavelength)

    rs = results["r_s"]
    rp = results["r_p"]

    psi = np.arctan(abs(rp / rs))
    delta = np.angle(1 / (-rp / rs)) + np.pi

    if reflect_delta:
        # Different ellipsometers / modelling software have different
        # conventions for what to do with Delta's above 180. WVASE appears to
        # reflect Delta around 180, which is what we have attempted to replicate.
        delta[delta > np.pi] = 2 * np.pi - delta[delta > np.pi]

    return psi * (180 / np.pi), delta * (180 / np.pi) + delta_offset


class ReflectModelSE:
    r"""
    Parameters
    ----------
    structure : refnx.reflect.Structure
        The interfacial structure.
    name : str, optional
        Name of the Model
    """

    def __init__(
        self,
        structure,
        delta_offset=0,
        name=None,
    ):
        if not isinstance(structure, StructureSE):
            raise ValueError("structure must be a StructureSE instance")

        self.name = name
        self._parameters = None
        self._flip_delta = False

        self.delta_offset = possibly_create_parameter(
            delta_offset, name="delta offset"
        )

        # to make it more like a refnx.analysis.Model
        self.fitfunc = None

        self._structure = None
        self.structure = structure

    def __call__(self, wavelength_aoi, p=None):
        r"""
        Calculate the generative model

        Parameters
        ----------
        wavelength_aoi : array-like
            An array of shape (N, 2) corresponding to the wavelengths (nm) and
            angle of incidences (deg) the ellipsometric measurements were
            performed at.
        p : refnx.analysis.Parameters, optional
            parameters required to calculate the model

        Returns
        -------
        psi, delta : np.ndarray
        """
        return self.model(wavelength_aoi, p=p)

    def __repr__(self):
        return (
            f"ReflectModel({self._structure!r}, name={self.name!r},"
            f" delta_offset = {self.delta_offset.value!r} "
        )

    def model(self, wavelength_aoi, p=None):
        r"""
        Calculate the ellipsometric values (psi, delta) of this model

        Parameters
        ----------
        wavelength_aoi : array-like
            An array of shape (N, 2) corresponding to the wavelengths (nm) and
            angle of incidences (deg) the ellipsometric measurements were
            performed at.
        p : refnx.analysis.Parameters, optional
            parameters required to calculate the model

        Returns
        -------
        reflectivity : np.ndarray
            Calculated reflectivity

        """
        if p is not None:
            self.parameters.pvals = np.array(p)

        wavelength, aois = wavelength_aoi.T
        psi = np.zeros_like(wavelength)
        delta = np.zeros_like(wavelength)

        unique_wavelengths = np.unique(wavelength)
        for wav in unique_wavelengths:
            idx = np.where(wavelength == wav)
            aoi = aois[idx]
            self.structure.wavelength = wav
            _psi, _delta = Delta_Psi_TMM(
                AOI=aoi,
                layers=self.structure.slabs()[..., :4],
                wavelength=wav,
                delta_offset=self.delta_offset.value,
                reflect_delta=self._flip_delta,
            )
            psi[idx] = _psi
            delta[idx] = _delta

        return psi, delta

    def logp(self):
        r"""
        Additional log-probability terms for the reflectivity model. Do not
        include log-probability terms for model parameters, these are
        automatically included elsewhere.

        Returns
        -------
        logp : float
            log-probability of structure.

        """
        return self.structure.logp()

    @property
    def structure(self):
        r"""
        :class:`refnx.reflect.Structure` - object describing the interface of
        a reflectometry sample.

        """
        return self._structure

    @structure.setter
    def structure(self, structure):
        if not isinstance(structure, StructureSE):
            raise ValueError("structure must be a StructureSE instance")

        self._structure = structure

        p = Parameters(name="instrument parameters")
        p.extend([self.delta_offset])

        self._parameters = Parameters(name=self.name)
        self._parameters.extend([p, structure.parameters])

    @property
    def parameters(self):
        r"""
        :class:`refnx.analysis.Parameters` - parameters associated with this
        model.

        """
        self.structure = self._structure
        return self._parameters



import os.path
from os.path import join as pjoin
import glob
import numpy as np
from numpy.testing import assert_allclose
from refnx._lib import flatten

import refellips
from refellips import (
    RI,
    DataSE,
    ReflectModelSE,
    ObjectiveSE,
    Cauchy,
    MixedSlabSE,
    load_material,
)


def test_cauchy_against_wvase():
    # Check the cauchy model behaves as expected
    A = 1.47
    B = 0.00495
    C = 0

    cauchy = Cauchy(A=A, B=B, C=C)

    pth = os.path.dirname(os.path.abspath(__file__))
    _f = pjoin(pth, "Cauchynk_fromWVASE.txt")
    wvase_output = np.loadtxt(_f)
    wavs = wvase_output[:, 0]
    refin = A + B / ((wavs / 1000) ** 2) + C / ((wavs / 1000) ** 4)

    refellips_RI = [cauchy.complex(wav).real for wav in wavs]

    assert_allclose(refellips_RI, wvase_output[:, 1], rtol=0.000001)
    assert_allclose(refellips_RI, refin, rtol=0.000001)


def test_RI_from_array():
    A = 1.47
    B = 0.00495
    C = 0

    wavs = np.arange(1, 101.0)
    ri_in = A + B / ((wavs) ** 2) + C / ((wavs) ** 4)
    ec_in = np.linspace(0.01, 0.2, 100)

    _f = RI((wavs, ri_in, ec_in))
    assert_allclose(_f._wav[0], 1000.0)
    ri = _f.complex(2000.0)
    assert_allclose(ri, complex(ri_in[1], ec_in[1]))


def test_dispersions_are_loadable():
    # test that all the bundled dispersion curves are loadable
    pth = os.path.dirname(os.path.abspath(refellips.__file__))
    pth = pjoin(pth, "materials")
    materials = glob.glob(pjoin(pth, "*.csv"))
    for material in materials:
        _f = RI(material)
        assert len(_f._wav) > 1


def test_mixedslab():
    a = load_material("water")
    b = Cauchy(A=1.47, B=0.00495)

    slab = MixedSlabSE(10, a, b, 0.5, 3)
    assert b.A in flatten(slab.parameters)
    assert slab.slabs().shape == (1, 5)
    assert_allclose(slab.vf_B.value, 0.5)

    a.wavelength = 400
    b.wavelength = 400
    riac = a.complex(400)
    ribc = b.complex(400)

    slab.vf_B.value = 0
    assert_allclose(slab.slabs()[0, 1], np.real(riac))
    assert_allclose(slab.slabs()[0, 2], np.imag(riac))

    slab.vf_B.value = 0.25
    overall = np.sqrt(0.75 * riac**2 + 0.25 * ribc**2)
    assert_allclose(slab.slabs()[0, 1], np.real(overall))
    assert_allclose(slab.slabs()[0, 2], np.imag(overall))



""""""
"""
BSD 3-Clause License

Copyright (c) 2020-2022,
Hayden Robertson (University of Newcastle)
Isaac Gresham (University of Sydney)
Andrew Nelson (ANSTO)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# -*- coding: utf-8 -*-
import numpy as np
import os
import os.path
import warnings
import glob

from refnx.reflect.structure import (
    Scatterer,
    Structure,
    Slab,
    Component,
    sld_profile,
)
from refnx.analysis import Parameters, Parameter, possibly_create_parameter
from refnx._lib import flatten
from refnx.reflect import _reflect as refcalc

# contracting the SLD profile can greatly speed a reflectivity calculation up.
contract_by_area = refcalc._contract_by_area


# list of material dispersion curves distributed with refellips
_pth = os.path.dirname(os.path.abspath(__file__))
_material_files = glob.glob(os.path.join(_pth, "materials/*.csv"))
materials = [os.path.basename(m)[:-4] for m in _material_files]


class ScattererSE(Scatterer):
    """
    Abstract base class for something that will have a refractive index.
    Inherited from refnx.reflect.structure.Scatterer
    """

    def __init__(self, name="", wavelength=None):
        self.name = name
        # by default energy dispersive scatterers for ellipsometry are energy
        # dispersive
        self.dispersive = True
        self.wavelength = wavelength

    def __str__(self):
        ri = complex(self)
        return f"n: {ri.real}, k: {ri.imag}"

    def __complex__(self):
        """
        The refractive index and extinction coefficient
        """
        return self.complex(None)

    def complex(self, wavelength):
        raise NotImplementedError(
            "The complex method is not implemented for this subclass"
        )

    def __call__(self, thick=0, rough=0, vfsolv=0):
        """
        Create a :class:`SlabSE`.
        Parameters
        ----------
        thick: refnx.analysis.Parameter or float
            Thickness of slab in Angstrom
        rough: refnx.analysis.Parameter or float
            Roughness of slab in Angstrom
        vfsolv: refnx.analysis.Parameter or float
            Volume fraction of water in slab
        Returns
        -------
        slab : SlabSE
            The newly made Slab.
        Example
        --------
        >>> # a RI object representing Silicon Dioxide using a given refractive
            index (e.g., n = 1.46) and extinction coefficent (e.g., k = 0).
        >>> sio2 = RI(dispersion=[1.46, 0], name='SiO2')
        >>> # create a Slab of SiO2 20 A in thickness, with a 3 A roughness
        >>> sio2_layer = sio2(20, 3)
        """
        return SlabSE(thick, self, rough, name=self.name, vfsolv=vfsolv)


class RI(ScattererSE):
    """
    Object representing a materials wavelength-dependent refractive index.

    An issue is that optical parameters are supplied in units of micrometers
    ('cause thats what seems to be used in refractive index repos and
    cauchy models), the wavelength of the incident radiation is supplied in
    nanometers (thats typical) and the fitting is done in angstroms. Very
    unpleasant.

    Parameters
    ----------
    dispersion : {str, tuple, np.ndarray)
        If a string then a dispersion curve will be loaded from a file that
        the string points to. The file is assumed to be of CSV format, with the
        first column holding the wavelength (in *microns*), with the second
        column specifying the refractive index. An optional third column can be
        present that should hold the extinction coefficient.

        If  `dispersion` has length 2 (float, float), then dispersion[0] points
        to the refractive index of the material and dispersion[1] points to the
        extinction coefficient. This refractive index is assumed to be
        wavelength independent.

        If `dispersion` has length 3, then dispersion[0], dispersion[1],
        dispersion[2] are assumed to hold arrays specifying the wavelength (in
        *microns*), refractive index, and extinction coefficient.
    wavelength : float
        default wavelength for calculation (nm)
    name : str, optional
        Name of material.

    Notes
    -----
    An RI object can be used to create a Slab
    """

    def __init__(self, dispersion=None, wavelength=658, name=""):
        super(RI, self).__init__(name=name, wavelength=wavelength)

        # _wav is only set if a wavelength dependent dispersion curve is loaded
        # assumed to be in nm
        self._wav = None
        self._RI = None
        self._EC = None

        if dispersion is None:
            raise RuntimeError("dispersion must be specified")

        if dispersion is not None:
            if type(dispersion) is str:
                if not len(name):
                    # if there is no name get it from the path
                    name = os.path.basename(dispersion).split(".")[0]

                vals = np.loadtxt(
                    dispersion, skiprows=1, delimiter=",", encoding="utf8"
                ).T
                self._wav = vals[0]
                self._RI = vals[1]
                self._EC = np.zeros_like(self._wav)
                if len(vals) == 3:
                    self._EC = vals[2]
                # convert wavelength from um to nm
                self._wav = self._wav * 1000
            elif len(dispersion) == 2:
                self._RI, self._EC = dispersion
            elif len(dispersion) == 3:
                # this is if you have an (3, N) array or tuple specifying
                # wavelength, RI, extinction coef.
                # wavelength assumed to be in *nm*
                self._wav, self._RI, self._EC = dispersion
                self._wav *= 1000
            else:
                raise TypeError("format not recognised")

    @property
    def parameters(self):
        return Parameters(name=self.name)

    def complex(self, wavelength):
        """
        Calculate a complex RI

        Parameters
        ----------
        wavelength : float
            wavelength of light in nm

        Returns
        -------
        RI : complex
            refractive index and extinction coefficient
        """
        wav = self.wavelength
        if np.any(wavelength):
            wav = wavelength

        if np.any(self._wav):
            # return a wavelength from a dispersion curve
            # TODO - raise a warning if the wavelength supplied is outside the
            # wavelength range covered by the data file.
            ri_real = np.interp(wav, self._wav, self._RI)
            ri_imag = np.interp(wav, self._wav, self._EC)
            return ri_real + 1j * ri_imag

        else:
            return self._RI + 1j * self._EC


def load_material(material):
    """
    Loads a dispersion curve from a file distributed with refellips.

    Parameters
    ----------
    material: str
        One of the materials in ``refellips.materials``

    Returns
    -------
    ri: refellips.RI

    Notes
    -----
    To get a list of the dispersion curves distributed with refellips examine
    the entries in ``refellips.materials``.
    """
    if material in materials:
        pth = os.path.join(_pth, "materials", f"{material}.csv")
        return RI(dispersion=pth)
    raise ValueError(f"{material} is not in the list of dispersion curves")


class Cauchy(ScattererSE):
    """
    Cauchy model for wavelength-dependent refractive index.

    Optical parameters are supplied in units of micrometers
    ('cause thats what seems to be used in refractive index repos and
    cauchy models), the wavelength of the incident radiation is supplied in
    nanometers (that's typical) and the fitting is done in angstroms.

    The refractive index is calculated as:
    ``A + (B * 1000**2) / (wav**2) + (C * 1000**4) / (wav**4)``

    where the factors of 1000 convert from microns to nm.

    Parameters
    ----------
    A : float or parameter
        Cauchy parameter A.
    B : float or parameter
        Cauchy parameter B in um^2. Default 0.
    C : float or parameter
        Cauchy parameter C in um^4. Default 0.
    wavelength : float
        default wavelength for calculation (nm)
    name : str, optional
        Name of material.
    """

    def __init__(self, A, B=0, C=0, wavelength=658, name=""):
        super().__init__(name=name, wavelength=wavelength)
        self.A = possibly_create_parameter(A, name=f"{name} - cauchy A")
        self.B = possibly_create_parameter(B, name=f"{name} - cauchy B")
        self.C = possibly_create_parameter(C, name=f"{name} - cauchy C")
        self._parameters = Parameters(name=name)
        self._parameters.extend([self.A, self.B, self.C])

    @property
    def parameters(self):
        return self._parameters

    def complex(self, wavelength):
        """
        Calculate a complex RI

        Parameters
        ----------
        wavelength : float
            wavelength of light in nm

        Returns
        -------
        RI : complex
            refractive index and extinction coefficient
        """
        wav = self.wavelength
        if np.any(wavelength):
            wav = wavelength

        real = (
            self.A.value
            + (self.B.value * 1000**2) / (wav**2)
            + (self.C.value * 1000**4) / (wav**4)
        )
        return real + 1j * 0.0


class ComponentSE(Component):
    """
    A base class for describing the structure of a subset of an interface.

    Parameters
    ----------
    name : str, optional
        The name associated with the Component

    Notes
    -----
    By setting the `Component.interfaces` property one can control the
    type of interfacial roughness between all the layers of an interfacial
    profile.
    """

    def __or__(self, other):
        """
        OR'ing SlabSE can create a :class:`StructureSE`.

        Parameters
        ----------
        other: refellips.StructureSE, refellips.ComponentSE
            Combines with this component to make a Structure

        Returns
        -------
        s: refellips.StructureSE
            The created Structure

        """
        # c = self | other
        p = StructureSE()
        p |= self
        p |= other
        return p

    def __mul__(self, n):
        """
        MUL'ing components makes them repeat.

        Parameters
        ----------
        n: int
            How many times you want to repeat the Component

        Returns
        -------
        s: refellipse.StructureSE
            The created Structure
        """
        # convert to integer, should raise an error if there's a problem
        n = operator.index(n)
        if n < 1:
            return StructureSE()
        elif n == 1:
            return self
        else:
            s = StructureSE()
            s.extend([self] * n)
            return s


class SlabSE(ComponentSE):
    """
    A slab component has uniform SLD over its thickness
    from refnx.reflect.structure.Slab
    Parameters
    ----------
    thick : refnx.analysis.Parameter or float
        thickness of slab (Angstrom)
    ri : :class:`refellips.ScattererSE`, complex, or float
        (complex) RI of film
    rough : refnx.analysis.Parameter or float
        roughness on top of this slab (Angstrom)
    name : str
        Name of this slab
    vfsolv : refnx.analysis.Parameter or float
        Volume fraction of solvent [0, 1]
    interface : {:class:`Interface`, None}, optional
        The type of interfacial roughness associated with the Slab.
        If `None`, then the default interfacial roughness is an Error
        function (also known as Gaussian roughness).
    """

    def __init__(self, thick, ri, rough, name="", vfsolv=0, interface=None):
        self.name = name
        self._interfaces = None
        self.thick = possibly_create_parameter(
            thick, name=f"{name} - thick", units="Å"
        )
        if isinstance(ri, ScattererSE):
            self.ri = ri
        else:
            self.ri = RI(ri)
        self.rough = possibly_create_parameter(
            rough, name=f"{name} - rough", units="Å"
        )
        self.vfsolv = possibly_create_parameter(
            vfsolv, name=f"{name} - volfrac solvent", bounds=(0.0, 1.0)
        )

        p = Parameters(name=self.name)
        p.extend([self.thick])
        p.extend(self.ri.parameters)
        p.extend([self.rough, self.vfsolv])

        self._parameters = p
        self.interfaces = interface

    def __repr__(self):
        return (
            f"SlabSE({self.thick!r}, {self.ri!r}, {self.rough!r},"
            f" name={self.name!r}, vfsolv={self.vfsolv!r},"
            f" interface={self.interfaces!r})"
        )

    @property
    def parameters(self):
        """
        :class:`refnx.analysis.Parameters` associated with this component
        """
        self._parameters.name = self.name
        return self._parameters

    def slabs(self, structure=None):
        """
        Slab representation of this component. See :class:`Component.slabs`
        """
        # speculative shortcut to prevent a number of attribute retrievals
        if self.ri.dispersive:
            ric = self.ri.complex(getattr(structure, "wavelength", None))
        else:
            ric = complex(self.ri)

        return np.array(
            [
                [
                    self.thick.value,
                    ric.real,
                    ric.imag,
                    self.rough.value,
                    self.vfsolv.value,
                ]
            ],
            dtype=float,
        )


class MixedSlabSE(ComponentSE):
    """
    A slab component made of two materials.

    Parameters
    ----------
    thick : refnx.analysis.Parameter or float
        thickness of slab (Angstrom)
    ri_A : ScattererSE
        refractive index of first material
    ri_B : ScattererSE
        refractive index of second material
    vf_B : float
        volume fraction of B in the layer. Volume fraction of A is calculated
        as ``1 - vf_B``.
    rough : refnx.analysis.Parameter or float
        roughness on top of this slab (Angstrom)
    name : str
        Name of this slab
    interface : {:class:`Interface`, None}, optional
        The type of interfacial roughness associated with the Slab.
        If `None`, then the default interfacial roughness is an Error
        function (also known as Gaussian roughness).
    """

    def __init__(
        self,
        thick,
        ri_A,
        ri_B,
        vf_B,
        rough,
        name="",
        interface=None,
    ):
        super().__init__(name=name)
        self.thick = possibly_create_parameter(
            thick, name=f"{name} - thick", units="Å"
        )
        self.ri_A = ri_A
        self.ri_B = ri_B
        self.vf_B = possibly_create_parameter(
            vf_B, name=f"{name} - vf_B", bounds=(0, 1)
        )

        self.rough = possibly_create_parameter(
            rough, name=f"{name} - rough", units="Å"
        )

        p = Parameters(name=self.name)
        p.append(self.thick)
        p.extend(self.ri_A.parameters)
        p.extend(self.ri_B.parameters)
        p.append(self.vf_B)
        p.append(self.rough)
        self._parameters = p
        self.interfaces = interface

    def __repr__(self):
        return (
            f"MixedSlabSE({self.thick!r}, {self.ri_A!r}, {self.ri_B!r},"
            f" {self.vf_B!r}, {self.rough!r}, name={self.name!r},"
            f" interface={self.interfaces!r})"
        )

    def __str__(self):
        return str(self.parameters)

    @property
    def parameters(self):
        """
        :class:`refnx.analysis.Parameters` associated with this component

        """
        self._parameters.name = self.name
        return self._parameters

    def slabs(self, structure=None):
        """
        Slab representation of this component. See :class:`Component.slabs`
        """
        wavelength = getattr(structure, "wavelength", None)
        if self.ri_A.dispersive:
            riac = self.ri_A.complex(wavelength)
        else:
            riac = complex(self.ri_A)

        if self.ri_B.dispersive:
            ribc = self.ri_B.complex(wavelength)
        else:
            ribc = complex(self.ri_B)

        ema = getattr(structure, "ema", "linear")
        dp = getattr(structure, "depolarisation_factor", 1 / 3)

        slabs = np.zeros((1, 5))
        N_avg = overall_ri(
            riac, ribc, vf_B=self.vf_B.value, ema=ema, depolarisation_factor=dp
        )
        slabs[0, 0] = self.thick.value
        slabs[0, 1] = np.real(N_avg)
        slabs[0, 2] = np.imag(N_avg)
        slabs[0, 3] = self.rough.value
        return slabs


class StructureSE(Structure):
    """
    Represents the interfacial Structure of an Ellipsometry sample.
    Successive Components are added to the Structure to construct the
    interface.

    Parameters
    ----------
    components : sequence
        A sequence of ComponentSE to initialise the Structure.
    name : str
        Name of this structure
    solvent : ScattererSE
        Specifies the refractive index of the solvent used for solvation. If no
        solvent is specified then the RI of the solvent is assumed to be
        the RI of `Structure[-1].slabs()[-1]` (after any possible slab order
        reversal).
    reverse_structure : bool
        If `StructureSE.reverse_structure` is `True` then the slab
        representation produced by `StructureSE.slabs` is reversed. The sld
        profile and calculated reflectivity will correspond to this
        reversed structure.
    contract : float
        If contract > 0 then an attempt to contract/shrink the slab
        representation is made. Use larger values for coarser
        profiles (and vice versa). A typical starting value to try might
        be 1.0.
    wavelength : float, None
        Wavelength the sample was measured at.
    ema : {'linear', 'maxwell-garnett', 'bruggeman'}
        Specifies the effective medium approximation for how the RI of a
        Component is mixed with the RI of the solvent. Further details
        regarding mixing are explained in the `slabs` method.
    depolarisation_factor : float, int
        The depolarisation factor is used only in the EMA calculations for
        the Maxwell-Garnett and Bruggeman methods. It describes the
        electric field screening: 0 prescribing no screening and 1
        prescribing maximum screening.
    """

    def __init__(
        self,
        components=(),
        name="",
        solvent=None,
        reverse_structure=False,
        contract=0,
        wavelength=None,
        ema="linear",
        depolarisation_factor=1 / 3,
    ):
        super().__init__()
        self._name = name
        self.solvent = solvent
        self._reverse_structure = bool(reverse_structure)

        #: **float** if contract > 0 then an attempt to contract/shrink the
        #: slab representation is made. Use larger values for coarser profiles
        #: (and vice versa). A typical starting value to try might be 1.0.
        self.contract = contract

        # used for energy dispersive measurements.
        self.wavelength = wavelength

        # if you provide a list of components to start with, then initialise
        # the structure from that
        self.data = [c for c in components if isinstance(c, ComponentSE)]

        self.ema = ema
        self._depolarisation_factor = depolarisation_factor

    sld_profile = None

    def __copy__(self):
        s = StructureSE(name=self.name, solvent=self._solvent)
        s.data = self.data.copy()
        return s

    @property
    def solvent(self):
        if self._solvent is None:
            if not self.reverse_structure:
                solv_slab = self[-1].slabs(self)
            else:
                solv_slab = self[0].slabs(self)
            return RI(complex(solv_slab[-1, 1], solv_slab[-1, 2]))
        else:
            return self._solvent

    @solvent.setter
    def solvent(self, ri):
        if ri is None:
            self._solvent = None
        elif isinstance(ri, ScattererSE):
            # don't make a new SLD object, use its reference
            self._solvent = ri
        else:
            solv = RI(ri)
            self._solvent = solv

    @property
    def depolarisation_factor(self):
        """ """
        return self._depolarisation_factor

    @depolarisation_factor.setter
    def depolarisation_factor(self, value):
        if 0 <= float(value) <= 1:
            self._depolarisation_factor = float(value)
        else:
            raise ValueError(
                "Depolarisation factor needs to be a float in [0, 1]."
            )

    def append(self, item):
        """
        Append a :class:`Component` to the Structure.

        Parameters
        ----------
        item: refnx.reflect.Component
            The component to be added.
        """
        if isinstance(item, ScattererSE):
            self.append(item())
            return

        if not isinstance(item, ComponentSE):
            raise ValueError(
                "You can only add ComponentSE objects to a structure"
            )
        super().append(item)

    def slabs(self, **kwds):
        r"""
        The slab representation of this structure.

        Returns
        -------
        slabs : :class:`np.ndarray`
            Slab representation of this structure.
            Has shape (N, 5).

            - slab[N, 0]
               thickness of layer N
            - slab[N, 1]
               *overall* RI.real of layer N (material AND solvent)
            - slab[N, 2]
               *overall* RI.imag of layer N (material AND solvent)
            - slab[N, 3]
               roughness between layer N and N-1
            - slab[N, 4]
               volume fraction of solvent in layer N.

        Notes
        -----
        If `Structure.reversed is True` then the slab representation order is
        reversed. The slab order is reversed before the solvation calculation
        is done. I.e. if `Structure.solvent == 'backing'` and
        `Structure.reversed is True` then the material that solvates the system
        is the component in `Structure[0]`, which corresponds to
        `Structure.slab[-1]`.

        Users can simulate mixing between two adjacent layers by specifying a
        volume fraction of solvent (`vfsolv`). The `overall_ri` function then
        performs the EMA using the specified method: 'linear',
        'maxwell-garnett' or 'bruggeman'. All EMA calculations are performed
        by using the complex dielectric function (i.e., square of refractive
        index and extinction coefficient).
        For a host layer (e_h) with volume fraction (vf) of impurities (e_i),
        the overall RI is calculated by

        >>> StructureSE.ema = 'linear'
        e_linear = e_h * (1 - vf) + e_i * vf

        >>> StructureSE.ema = 'maxwell-garnett'
        >>> StructureSE.depolarisation_factor = 1/3
        top = e_h + (depolarisation_factor * (1 - vf) + vf) * (e_i - e_h)
        bottom = e_h + depolarisation_factor * (1 - vf) * (e_i - e_h)
        e_MG = e_h * top_r / bottom_r

        >>> StructureSE.ema = 'bruggeman'
        >>> StructureSE.depolarisation_factor = 1/3
        b = e_h * ((1 - vf) - depolarisation_factor) + e_i * (vf - depolarisation_factor)
        e_BG = (b + np.sqrt(b**2 - 4 * (depolarisation_factor - 1) *
                                 (vf * e_h * e_i * depolarisation_factor
                                  ))) / (2 * (1 - depolarisation_factor))
        """

        if not len(self):
            return None

        if not (
            isinstance(self.data[-1], (SlabSE))
            and isinstance(self.data[0], (SlabSE))
        ):
            raise ValueError(
                "The first and last Components in a StructureSE need to be SlabsSE"
            )
        # over-ride the wavelength
        if "wavelength" in kwds:
            self.wavelength = float(kwds["wavelength"])

        # Each layer can be given a different type of roughness profile
        # that defines transition between successive layers.
        # The default interface is specified by None (= Gaussian roughness)
        interfaces = flatten(self.interfaces)
        if all([i is None for i in interfaces]):
            # if all the interfaces are Gaussian, then simply concatenate
            # the default slabs property of each component.
            sl = [c.slabs(structure=self) for c in self.components]

            try:
                slabs = np.concatenate(sl)
            except ValueError:
                # some of slabs may be None. np can't concatenate arr and None
                slabs = np.concatenate([s for s in sl if s is not None])
        else:
            # there is a non-default interfacial roughness, create a microslab
            # representation
            slabs = self._micro_slabs()

        # if the slab representation needs to be reversed.
        reverse = self.reverse_structure
        if reverse:
            roughnesses = slabs[1:, 3]
            slabs = np.flipud(slabs)
            slabs[1:, 3] = roughnesses[::-1]
            slabs[0, 3] = 0.0

        if (slabs[:, 4] > 0).any():
            # overall SLD is a weighted average of the vfs and slds
            # accessing self.solvent leads to overhead from object
            # creation.
            if self._solvent is not None:
                solv = self._solvent
            else:
                # we should always choose the solvating material to be the last
                # slab. If the structure is not reversed then you want the last
                # slab. If the structure is reversed then you should want to
                # use the first slab, but the code block above reverses the
                # slab order, so we still want the last one
                solv = complex(slabs[-1, 1], slabs[-1, 2])

            slabs[1:-1] = self.overall_ri(slabs[1:-1], solv)

        if self.contract > 0:
            return contract_by_area(slabs, self.contract)
        else:
            return slabs

    def ri_profile(self, z=None, align=0, max_delta_z=None):
        """
        Calculates an RI profile, as a function of distance through the
        interface.

        Parameters
        ----------
        z : float
            Interfacial distance (Angstrom) measured from interface between the
            fronting medium and the first layer.
        align: int, optional
            Places a specified interface in the slab representation of a
            Structure at z = 0. Python indexing is allowed, e.g. supplying -1
            will place the backing medium at z = 0.
        max_delta_z : {None, float}, optional
            If specified this will control the maximum spacing between SLD
            points. Only used if `z is None`.

        Returns
        -------
        ri : float
            refractive index

        Notes
        -----
        This can be called in vectorised fashion.
        """
        slabs = self.slabs()
        if (
            (slabs is None)
            or (len(slabs) < 2)
            or (not isinstance(self.data[0], SlabSE))
            or (not isinstance(self.data[-1], SlabSE))
        ):
            raise ValueError(
                "Structure requires fronting and backing"
                " Slabs in order to calculate."
            )

        zed, sld = sld_profile(slabs, z=z, max_delta_z=max_delta_z)

        offset = 0
        if align != 0:
            align = int(align)
            if align >= len(slabs) - 1 or align < -1 * len(slabs):
                raise RuntimeError(
                    "abs(align) has to be less than " "len(slabs) - 1"
                )
            # to figure out the offset you need to know the cumulative distance
            # to the interface
            slabs[0, 0] = slabs[-1, 0] = 0.0
            if align >= 0:
                offset = np.sum(slabs[: align + 1, 0])
            else:
                offset = np.sum(slabs[:align, 0])

        return zed - offset, sld

    def plot(self, pvals=None, samples=0, fig=None, align=0):
        """
        Plot the structure.

        Requires matplotlib be installed.

        Parameters
        ----------
        pvals : np.ndarray, optional
            Numeric values for the Parameter's that are varying
        samples: number
            If this structures constituent parameters have been sampled, how
            many samples you wish to plot on the graph.
        fig: Figure instance, optional
            If `fig` is not supplied then a new figure is created. Otherwise
            the graph is created on the current axes on the supplied figure.
        align: int, optional
            Aligns the plotted structures around a specified interface in the
            slab representation of a Structure. This interface will appear at
            z = 0 in the sld plot. Note that Components can consist of more
            than a single slab, so some thought is required if the interface to
            be aligned around lies in the middle of a Component. Python
            indexing is allowed, e.g. supplying -1 will align at the backing
            medium.

        Returns
        -------
        fig, ax : :class:`matplotlib.Figure`, :class:`matplotlib.Axes`
          `matplotlib` figure and axes objects.

        """
        import matplotlib.pyplot as plt

        params = self.parameters

        if pvals is not None:
            params.pvals = pvals

        if fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            ax = fig.gca()

        if samples > 0:
            saved_params = np.array(params)
            # Get a number of chains, chosen randomly, and plot the model.
            for pvec in self.parameters.pgen(ngen=samples):
                params.pvals = pvec

                ax.plot(*self.ri_profile(align=align), color="k", alpha=0.01)

            # put back saved_params
            params.pvals = saved_params

        ax.plot(*self.sld_profile(align=align), color="red", zorder=20)
        ax.set_ylabel("RI$")
        ax.set_xlabel("z / $\\AA$")

        return fig, ax

    def __ior__(self, other):
        """
        Build a structure by `IOR`'ing Structures/Components/SLDs.

        Parameters
        ----------
        other: :class:`StructureSE`, :class:`ComponentSE`, :class:`RI`
            The object to add to the structure.
        """
        # self |= other
        if isinstance(other, ComponentSE):
            self.append(other)
        elif isinstance(other, StructureSE):
            self.extend(other.data)
        elif isinstance(other, ScattererSE):
            slab = other(0, 0)
            self.append(slab)
        else:
            raise ValueError()

        return self

    def __or__(self, other):
        """
        Build a structure by `OR`'ing Structures/Components/SLDs.

        Parameters
        ----------
        other: :class:`StructureSE`, :class:`ComponentSE`, :class:`SLD`
            The object to add to the structure.

        Examples
        --------

        >>> air = RI(dispersion=[1,0], name='air')
        >>> sio2 = RI(dispersion=[1.46, 0], name='SiO2')
        >>> si = RI(dispersion=[3.84, 0.016], name='Si')
        >>> structure = StructureSE()
        >>> structure = air() | sio2(20) | si()
        """
        # c = self | other
        p = StructureSE()
        p |= self
        p |= other
        return p

    def overall_ri(self, slabs, solvent):
        """
        Calculates the overall refractive index of the material and solvent RI
        in a layer.

        Parameters
        ----------
        slabs : np.ndarray
            Slab representation of the layers to be averaged.
        solvent : complex or ScattererSE
            RI of solvating material.

        Returns
        -------
        averaged_slabs : np.ndarray
            the averaged slabs.

        """
        solv = solvent
        if isinstance(solvent, ScattererSE):
            solv = solvent.complex(self.wavelength)

        vf = slabs[..., 4]
        N = slabs[..., 1] + slabs[..., 2] * 1j

        N_avg = overall_ri(
            N,
            solv,
            vf_B=vf,
            ema=self.ema,
            depolarisation_factor=self.depolarisation_factor,
        )
        slabs[..., 1] = np.real(N_avg)
        slabs[..., 2] = np.imag(N_avg)

        return slabs


def overall_ri(ri_A, ri_B, vf_B=0.0, ema="linear", depolarisation_factor=1 / 3):
    """
    Calculates the overall refractive index of two materials.

    Parameters
    ----------
    ri_A: complex, array-like
        RI of material A
    ri_B: complex
        RI of material B
    vf_B: float, optional
        volume fraction of material B. The volume fraction of A is calculated
        as ``1 - vf_B``.
    ema : {'linear', 'maxwell-garnett', 'bruggeman'}
        Specifies how refractive indices are mixed together.
    depolarisation_factor : float, optional
        Depolarisation factor. Default is 1/3.

    Returns
    -------
    ri_avg : complex
        the averaged material RI
    """
    # E is the complex dielectric function
    E_a = np.power(ri_A, 2)
    E_b = np.power(ri_B, 2)

    if ema == "linear":
        E_avg = (1 - vf_B) * E_a + vf_B * E_b

    elif ema == "maxwell-garnett":
        top = E_a + (depolarisation_factor * (1 - vf_B) + vf_B) * (E_b - E_a)
        bottom = E_a + depolarisation_factor * (1 - vf_B) * (E_b - E_a)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            v = top / bottom
            v = np.where(np.isfinite(v), v, 0)
            E_avg = E_a * v

    elif ema == "bruggeman":
        # The solution to the Bruggeman EMA method is solved using the
        # quadratic equation, only one of which is physically reasonable.

        b = E_a * ((1 - vf_B) - depolarisation_factor) + E_b * (
            vf_B - depolarisation_factor
        )

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            v = (
                b
                + np.sqrt(
                    b**2
                    - 4
                    * (depolarisation_factor - 1)
                    * (E_a * E_b * depolarisation_factor)
                )
            ) / (2 * (1 - depolarisation_factor))
            E_avg = np.where(np.isfinite(v), v, 0)
    else:
        raise RuntimeError("No other method of mixing is known")

    ri_avg = np.sqrt(E_avg)
    return ri_avg



# -*- coding: utf-8 -*-

import numpy as np
from numpy.linalg import LinAlgError

from scipy.optimize._numdiff import approx_derivative


from refnx.analysis.objective import BaseObjective
from .dataSE import DataSE

from refnx.analysis import (
    is_parameter,
    Parameter,
    possibly_create_parameter,
    is_parameters,
    Parameters,
    Interval,
    PDF,
)
from refnx._lib import unique as f_unique
from refnx._lib import flatten, approx_hess2


class ObjectiveSE(BaseObjective):
    """
    Objective function for using with curvefitters such as
    `refnx.analysis.curvefitter.CurveFitter`.

    Parameters
    ----------
    model : refnx.analysis.Model
        the residuals model function. One can also provide an object that
        inherits `refnx.analysis.Model`.
    data : refnx.dataset.Data1D
        data to be analysed.
    lnsigma : float or refnx.analysis.Parameter, optional
        Used if the  experimental uncertainty (`data.y_err`) underestimated by
        a constant fractional amount. The experimental uncertainty is modified
        as:

        `s_n**2 = y_err**2 + exp(lnsigma * 2) * model**2`

        See `Objective.logl` for more details.
    use_weights : bool
        use experimental uncertainty in calculation of residuals and
        logl, if available. If this is set to False, then you should also
        set `self.lnsigma.vary = False`, it will have no effect on the fit.
    transform : callable, optional
        the model, data and data uncertainty are transformed by this
        function before calculating the likelihood/residuals. Has the
        signature `transform(data.x, y, y_err=None)`, returning the tuple
        (`transformed_y, transformed_y_err`).
    logp_extra : callable, optional
        user specifiable log-probability term. This contribution is in
        addition to the log-prior term of the `model` parameters, and
        `model.logp`, as well as the log-likelihood of the `data`. Has
        signature:
        `logp_extra(model, data)`. The `model` will already possess
        updated parameters. Beware of including the same log-probability
        terms more than once.
    name : str
        Name for the objective.

    Notes
    -----
    For parallelisation `logp_extra` needs to be picklable.

    """

    def __init__(
        self,
        model,
        data,
        lnsigma=None,
        use_weights=True,
        transform=None,
        logp_extra=None,
        name=None,
    ):
        self.model = model
        # should be a DataSE instance
        if type(data) is DataSE:
            self.data = data
        else:
            print("bad")
            self.data = DataSE(data=data)

        self.lnsigma = lnsigma
        if lnsigma is not None:
            self.lnsigma = possibly_create_parameter(lnsigma, "lnsigma")

        self._use_weights = use_weights
        self.transform = transform
        self.logp_extra = logp_extra
        self.name = name
        if name is None:
            self.name = id(self)

        self.model._flip_delta = self.data._delta_flipped

    def __str__(self):
        s = ["{:_>80}".format("")]
        s.append("Objective - {0}".format(self.name))

        # dataset name
        if self.data.name is None:
            s.append("Dataset = {0}".format(self.data))
        else:
            s.append("Dataset = {0}".format(self.data.name))

        s.append("datapoints = {0}".format(self.npoints))
        s.append("chi2 = {0}".format(self.chisqr()))
        s.append("Weighted = {0}".format(self.weighted))
        s.append("Transform = {0}".format(self.transform))
        s.append(str(self.parameters))

        return "\n".join(s)

    def __repr__(self):
        return (
            "Objective({model!r}, {data!r},"
            " lnsigma={lnsigma!r},"
            " use_weights={_use_weights},"
            " transform={transform!r},"
            " logp_extra={logp_extra!r},"
            " name={name!r})".format(**self.__dict__)
        )

    @property
    def weighted(self):
        """
        **bool** Does the data have weights (`data.y_err`), and is the
        objective using them?

        """
        return self.data.weighted and self._use_weights

    @weighted.setter
    def weighted(self, use_weights):
        self._use_weights = bool(use_weights)

    @property
    def npoints(self):
        """
        **int** the number of points in the dataset.

        """
        return self.data._wavelength.size

    def varying_parameters(self):
        """
        Returns
        -------
        varying_parameters : refnx.analysis.Parameters
            The varying Parameter objects allowed to vary during the fit.

        """
        # create and return a Parameters object because it has the
        # __array__ method, which allows one to quickly get numerical values.
        p = Parameters()
        p.data = list(f_unique(p for p in flatten(self.parameters) if p.vary))
        return p

    def residuals(self, pvals=None):
        """
        Calculates the residuals for a given fitting system.

        Parameters
        ----------
        pvals : array-like or refnx.analysis.Parameters
            values for the varying or entire set of parameters

        Returns
        -------
        residuals : np.ndarray
            Residuals, `(data.y - model) / y_err`.

        """

        self.setp(pvals)

        wavelength, aoi, psi_d, delta_d = self.data.data
        wavelength_aoi = np.c_[wavelength, aoi]

        psi, delta = self.model(wavelength_aoi)
        return np.r_[psi - psi_d, delta - delta_d]

    def chisqr(self, pvals=None):
        """
        Calculates the chi-squared value for a given fitting system.

        Parameters
        ----------
        pvals : array-like or refnx.analysis.Parameters
            values for the varying or entire set of parameters

        Returns
        -------
        chisqr : np.ndarray
            Chi-squared value, `np.sum(residuals**2)`.

        """
        # TODO reduced chisqr? include z-scores for parameters? DOF?
        self.setp(pvals)
        res = self.residuals(None)
        return np.dot(res, res)

    @property
    def parameters(self):
        """
        :class:`refnx.analysis.Parameters`, all the Parameters contained in the
        fitting system.

        """
        if is_parameter(self.lnsigma):
            return self.lnsigma | self.model.parameters
        else:
            return self.model.parameters

    def setp(self, pvals):
        """
        Set the parameters from pvals.

        Parameters
        ----------
        pvals : array-like or refnx.analysis.Parameters
            values for the varying or entire set of parameters

        """
        if pvals is None:
            return

        # set here rather than delegating to a Parameters
        # object, because it may not necessarily be a
        # Parameters object
        _varying_parameters = self.varying_parameters()
        if len(pvals) == len(_varying_parameters):
            for idx, param in enumerate(_varying_parameters):
                param.value = pvals[idx]
            return

        # values supplied are enough to specify all parameter values
        # even those that are repeated
        flattened_parameters = list(flatten(self.parameters))
        if len(pvals) == len(flattened_parameters):
            for idx, param in enumerate(flattened_parameters):
                param.value = pvals[idx]
            return

        raise ValueError(
            f"Incorrect number of values supplied ({len(pvals)})"
            f", supply either the full number of parameters"
            f" ({len(flattened_parameters)}, or only the varying"
            f" parameters ({len(_varying_parameters)})."
        )

    def prior_transform(self, u):
        """
        Calculate the prior transform of the system.

        Transforms uniform random variates in the unit hypercube,
        `u ~ uniform[0.0, 1.0)`, to the parameter space of interest, according
        to the priors on the varying parameters.

        Parameters
        ----------
        u : array-like
            Size of the varying parameters

        Returns
        -------
        pvals : array-like
            Scaled parameter values

        Notes
        -----
        If a parameter has bounds, `x ~ Unif[-10, 10)` then the scaling from
        `u` to `x` is done as follows:

        .. code-block:: python

            x = 2. * u - 1.  # scale and shift to [-1., 1.)
            x *= 10.  # scale to [-10., 10.)

        """
        var_pars = self.varying_parameters()
        pvals = np.empty(len(var_pars), dtype=np.float64)

        for i, var_par in enumerate(var_pars):
            pvals[i] = var_par.bounds.invcdf(u[i])

        return pvals

    def logp(self, pvals=None):
        """
        Calculate the log-prior of the system.

        Parameters
        ----------
        pvals : array-like or refnx.analysis.Parameters
            values for the varying or entire set of parameters

        Returns
        -------
        logp : float
            log-prior probability

        Notes
        -----
        The log-prior is calculated as:

        .. code-block:: python

            logp = np.sum(param.logp() for param in
                             self.varying_parameters())

        """
        self.setp(pvals)

        logp = np.sum(
            [
                param.logp()
                for param in f_unique(
                    p for p in flatten(self.parameters) if p.vary
                )
            ]
        )

        if not np.isfinite(logp):
            return -np.inf

        return logp

    def logl(self, pvals=None):
        """
        Calculate the log-likelhood of the system.

        Parameters
        ----------
        pvals : array-like or refnx.analysis.Parameters
            values for the varying or entire set of parameters

        Returns
        -------
        logl : float
            log-likelihood probability

        Notes
        -----
        The log-likelihood is calculated as:

        .. code-block:: python

            logl = -0.5 * np.sum(((y - model) / s_n)**2
                                 + np.log(2 * pi * s_n**2))
            logp += self.model.logp()
            logp += self.logp_extra(self.model, self.data)

        where

        .. code-block:: python

            s_n**2 = y_err**2 + exp(2 * lnsigma) * model**2

        At the moment s_n**2, the variance of the measurement uncertainties,
        is assumed to be unity. A future release may implement those
        uncertainties
        """
        self.setp(pvals)
        wavelength, aoi, psi_d, delta_d = self.data.data
        wavelength_aoi = np.c_[wavelength, aoi]

        psi, delta = self.model(wavelength_aoi)

        model = np.r_[psi, delta]

        logl = 0.0

        # TODO investigate ellipsometry uncertainties
        # here just set it to unity
        y_err = 1
        if self.lnsigma is not None:
            var_y = (
                y_err * y_err + np.exp(2 * float(self.lnsigma)) * model * model
            )
        else:
            var_y = y_err**2

        # TODO do something sensible coz the data isn't weighted
        if self.weighted:
            logl += np.log(2 * np.pi * var_y)

        logl += (np.r_[psi_d, delta_d] - model) ** 2 / var_y

        # nans play havoc
        if np.isnan(logl).any():
            raise RuntimeError("ObjectiveSE.logl encountered a NaN.")

        return -0.5 * np.sum(logl)

    def logpost(self, pvals=None):
        """
        Calculate the log-probability of the curvefitting system

        Parameters
        ----------
        pvals : array-like or refnx.analysis.Parameters
            values for the varying or entire set of parameters

        Returns
        -------
        logpost : float
            log-probability

        Notes
        -----
        The overall log-probability is the sum of the log-prior and
        log-likelihood. The log-likelihood is not calculated if the log-prior
        is impossible (`logp == -np.inf`).

        """
        self.setp(pvals)
        logpost = self.logp()

        # only calculate the probability if the parameters have finite
        # log-prior
        if not np.isfinite(logpost):
            return -np.inf

        logpost += self.logl()
        return logpost

    def covar(self):
        """
        Estimates the covariance matrix of the curvefitting system.

        Returns
        -------
        covar : np.ndarray
            Covariance matrix

        """
        _pvals = np.array(self.varying_parameters())

        used_residuals_scaler = False

        def residuals_scaler(vals):
            return np.squeeze(self.residuals(_pvals * vals))

        try:
            # we should be able to calculate a Jacobian for a parameter whose
            # value is zero. However, the scaling approach won't work.
            # This will force Jacobian calculation by unscaled parameters
            if np.any(_pvals == 0):
                raise FloatingPointError()

            with np.errstate(invalid="raise"):
                jac = approx_derivative(residuals_scaler, np.ones_like(_pvals))
            used_residuals_scaler = True
        except FloatingPointError:
            jac = approx_derivative(self.residuals, _pvals)
        finally:
            # using approx_derivative changes the state of the objective
            # parameters have to make sure they're set at the end
            self.setp(_pvals)

        # need to create this because GlobalObjective does not have
        # access to all the datapoints being fitted.
        n_datapoints = np.size(jac, 0)

        # covar = J.T x J

        # from scipy.optimize.minpack.py
        # eliminates singular parameters
        _, s, VT = np.linalg.svd(jac, full_matrices=False)
        threshold = np.finfo(float).eps * max(jac.shape) * s[0]
        s = s[s > threshold]
        VT = VT[: s.size]
        covar = np.dot(VT.T / s**2, VT)

        if used_residuals_scaler:
            # unwind the scaling.
            covar = covar * np.atleast_2d(_pvals) * np.atleast_2d(_pvals).T

        pvar = np.diagonal(covar).copy()
        psingular = np.where(pvar == 0)[0]

        if len(psingular) > 0:
            var_params = self.varying_parameters()
            singular_params = [var_params[ps] for ps in psingular]

            raise LinAlgError(
                "The following Parameters have no effect on"
                " Objective.residuals, please consider fixing"
                " them.\n" + repr(singular_params)
            )

        scale = 1.0
        # scale by reduced chi2 if experimental uncertainties weren't used.
        if not (self.weighted):
            scale = self.chisqr() / (
                n_datapoints - len(self.varying_parameters())
            )

        return covar * scale

    def pgen(self, ngen=1000, nburn=0, nthin=1):
        """
        Yield random parameter vectors from the MCMC samples. The objective
        state is not altered.

        Parameters
        ----------
        ngen : int, optional
            the number of samples to yield. The actual number of samples
            yielded is `min(ngen, chain.size)`
        nburn : int, optional
            discard this many steps from the start of the chain
        nthin : int, optional
            only accept every `nthin` samples from the chain

        Yields
        ------
        pvec : np.ndarray
            A randomly chosen parameter vector

        """
        yield from self.parameters.pgen(ngen=ngen, nburn=nburn, nthin=nthin)

    def plot(self, xaxis=None, plot_labels=True, fig=None):
        """
        Plot the data/model.

        Requires matplotlib be installed.

        Parameters
        ----------
        xaxis : String, optional
            Either 'aoi' or 'wavelength'. If none specified, 'wavelength'
            will be chosen unless there is more than 1 unique aoi.
        plot_labels : Bool, optional
            Whether to plot axis labels. The default is True.
        fig: Figure instance, optional
            If `fig` is not supplied then a new figure is created. Otherwise
            the graph is created on the current axes on the supplied figure.

        Returns
        -------
        fig, ax : :class:`matplotlib.Figure`, :class:`matplotlib.Axes`
            `matplotlib` figure and axes objects.

        """

        data = self.data
        model = self.model

        if fig is None:
            import matplotlib.pyplot as plt

            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            ax = fig.gca()

        axt = ax.twinx()

        unique_wavs = np.unique(data.wavelength)
        unique_aois = np.unique(data.aoi)

        if xaxis is None:
            if len(unique_aois) > 1:
                xaxis = "aoi"
            else:
                xaxis = "wavelength"

        if xaxis == "aoi":
            aois = np.linspace(np.min(data.aoi), np.max(data.aoi))
            x = data.aoi
            xlab = "AOI (°)"

            for wav in unique_wavs:
                psis, deltas = model(np.c_[np.ones_like(aois) * wav, aois])
                ax.plot(aois, psis, color="r")
                axt.plot(aois, deltas, color="b")

        elif xaxis == "wavelength":
            wavs = np.linspace(np.min(data.wavelength), np.max(data.wavelength))
            x = data.wavelength
            xlab = "Wavelength (nm)"

            for aoi in unique_aois:
                psi, delta = model(np.c_[wavs, np.ones_like(wavs) * aoi])
                ax.plot(np.ones_like(psi) * wavs, psi, color="r")
                axt.plot(np.ones_like(delta) * wavs, delta, color="b")

        p = ax.scatter(x, data.psi, color="r")
        d = axt.scatter(x, data.delta, color="b")

        ax.legend(handles=[p, d], labels=["Psi", "Delta"])

        if plot_labels:
            ax.set(ylabel="Psi", xlabel=xlab)
            axt.set(ylabel="Delta")

        return fig, ax




# THIS FILE IS GENERATED FROM refellips SETUP.PY
short_version = '0.0.2'
version = '0.0.2'
full_version = '0.0.2'
git_revision = '46e54b6fb70bd47acc48f301c6af6c6af598bee7'
release = True
if not release:
    version = full_version



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import os
import subprocess
import sys


packages = find_packages()


# versioning
MAJOR = 0
MINOR = 0
MICRO = 2
ISRELEASED = True
VERSION = f"{MAJOR}.{MINOR}.{MICRO}"


# are we on windows, darwin, etc?
platform = sys.platform
packages = find_packages()


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = "refellips"

    def run_tests(self):
        import shlex
        import pytest

        print("Running tests with pytest")
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


info = {
    "packages": packages,
    "include_package_data": True,
    "cmdclass": {"test": PyTest},
}


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, env=env
        ).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(["git", "rev-parse", "HEAD"])
        GIT_REVISION = out.strip().decode("ascii")
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside
    # write_version_py(), otherwise the import of refellips.version messes
    # up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists(".git"):
        GIT_REVISION = git_version()
    elif os.path.exists("refellips/version.py"):
        # must be a source distribution, use existing version file
        # load it as a separate module to not load refellips/__init__.py
        import imp

        version = imp.load_source("refellips.version", "refellips/version.py")
        GIT_REVISION = version.git_revision
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += ".dev0+" + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename="refellips/version.py"):
    cnt = """
# THIS FILE IS GENERATED FROM refellips SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, "w")
    try:
        a.write(
            cnt
            % {
                "version": VERSION,
                "full_version": FULLVERSION,
                "git_revision": GIT_REVISION,
                "isrelease": str(ISRELEASED),
            }
        )
    finally:
        a.close()


def setup_package():
    # Rewrite the version file every time
    write_version_py()
    info["version"] = get_version_info()[0]
    info["zip_safe"] = False
    print(info["version"])

    setup(**info)


if __name__ == "__main__":
    setup_package()
